# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/06_learner.ipynb (unless otherwise specified).

__all__ = ['create_classifier', 'Learner', 'InterpretModel']

# Cell
import tensorflow as tf
import tensorflow_addons as tfa
from tensorflow.keras.models import Model
from .datagenerator import Dataset

from typeguard import check_argument_types, check_return_type

# Cell
from PIL import Image
import numpy as np
from tf_keras_vis.utils import normalize
from tf_keras_vis.gradcam import GradcamPlusPlus, Gradcam
from functools import partial

import matplotlib.pyplot as plt
import matplotlib.cm as cm

# Cell
def create_classifier(base_model_fn:callable, num_classes:int,
                      weights='imagenet', dropout=0,
                      include_top=False,
                      name=None):

    outputs = 1 if num_classes == 2 else num_classes

    base_model = base_model_fn(
        include_top=include_top,
        weights=weights,
    )
    if include_top: return base_model
    model = tf.keras.Sequential(name=name)
    model.add(base_model)
    model.add(tf.keras.layers.GlobalAveragePooling2D())
    model.add(tf.keras.layers.Dropout(dropout))
    model.add(tf.keras.layers.Dense(outputs, name='output'))

    return model

# Cell
class Learner(Model):
    """
    The Learner class inherits tf.keras.Model and contains everything a model needs for training.
    It exposes learner.cyclic_fit method which trains the model using Cyclic Learning rate discovered by Leslie Smith.

    Arguments:
    ds: Dataset object
    base_model: callable function from keras.applications

    """
    _AUTOTUNE = tf.data.experimental.AUTOTUNE

    def __init__(self, ds: Dataset, base_model_fn:callable, pretrained:bool=True, include_top=False, **kwargs):
        assert check_argument_types()

        super(Learner, self).__init__()
        self.ds = ds
        self.total = len(ds)
        self.NUM_CLASSES = ds.NUM_CLASSES
        self.gradcam = None
        self.include_top = include_top

        weights = 'imagenet' if pretrained else None

        #self.base_model = base_model_fn.name

        self.model = create_classifier(
            base_model_fn,
            self.NUM_CLASSES,
            dropout=kwargs.get('dropout', 0.5),
            include_top=include_top,
            name=kwargs.get('name', None)
        )


    def build(self): pass

    def summary(self): return self.model.summary()

    #def get_layer(name=None, index=None): return self.model(name, index)

    def compile(self, *args, **kwargs): return self.model.compile(*args, **kwargs)

    def call(self, *args, **kwargs): return self.model.call(*args, **kwargs)

    def fit(self, *args, **kwargs): return self.model.fit(*args, **kwargs)

    def warmup(self):pass

    def prewhiten(self, image):
        image = tf.cast(image, tf.float32)
        image = image / 127.5 - 1.0
        return image

    def rescale(self, image, label):
        image = self.prewhiten(image)
        return image, label

    def _get_optimizer(self,
                       optimizer,
                       momentum=0.9,
                       **kwargs
                      ):
        if optimizer.__name__=='SGD':
            optimizer = partial(optimizer,
                momentum=momentum,
                nesterov=kwargs.get('nesterov', True)
            )
        else:
            optimizer = partial(optimizer,
                momentum=momentum,
            )
        return optimizer


    def _prepare_dl(self, bs=8, shuffle=True):
        ds = self.ds
        dl = ds.get_tf_dataset(shuffle=shuffle)
        dl = dl.map(self.rescale, Learner._AUTOTUNE)
        return dl.batch(bs).prefetch(Learner._AUTOTUNE)


    def cyclic_fit(self,
                   epochs,
                   batch_size,
                   lr_range=(1e-4, 1e-2),
                   optimizer=tf.keras.optimizers.SGD,
                   momentum=0.9,
                   validation_data=None,
                   callbacks=None,
                   *args,
                   **kwargs
                  ):
        """Trains model on ds as train data with cyclic learning rate.
        Dataset will be automatically converted into `tf.data` format and images will be prewhitened in range of [-1, 1].
        """

        self.max_lr, self.min_lr = lr_range
        ds = self.ds

        step_size = 2 * len(self.ds)//batch_size

        lr_schedule = tfa.optimizers.Triangular2CyclicalLearningRate(
                                    initial_learning_rate=lr_range[0],
                                    maximal_learning_rate=lr_range[1],
                                    step_size=kwargs.get('step_size', step_size),
                                    scale_mode=kwargs.get('scale', 'cycle'))


        optimizer = self._get_optimizer(optimizer, momentum=momentum)
        optimizer = optimizer(learning_rate=lr_schedule)
        self.model.optimizer = optimizer

        return self.model.fit(
            self._prepare_dl(batch_size, kwargs.get('shuffle', True)),
            validation_data=validation_data,
            epochs=epochs,
            callbacks=callbacks
        )

    def gradcam(self, image:Image.Image, image_size=None):
        # assert check_argument_types()

        def model_modifier(m):
            """Converts sigmoid to linear
            """
            m.layers[-1].activation = tf.keras.activations.linear
            return m


        if image_size:
            image_size = self.ds.img_sz_list.get_size()
        image = image.resize(image_size)

        X = np.asarray(image, np.float32)
        X = self.prewhiten(X)
        X = np.expand_dims(X, 0)

        if self.gradcam is None:
            self.gradcam = Gradcam(self.model,
                              model_modifier,
                              clone=True)

        cam = gradcam(get_loss,
              X,
              penultimate_layer=-1, # model.layers number
              seek_penultimate_conv_layer=False,

             )


# Cell
class InterpretModel():
    def __init__(self, gradcam_pp:bool, learner:Learner, clone:bool=False):
        """Args:
        gradcam_pp: if True GradCam class will be used else GradCamPlusplus
        clone: whether GradCam will clone learner.model
        """
        if gradcam_pp:
            self.gradcam_fn = GradcamPlusPlus
        else:
            self.gradcam_fn = Gradcam
        self.learner = learner

        self.gradcam = self.gradcam_fn(learner.model,
                                      self.model_modifier,
                                      clone=clone)

        if self.learner.include_top is not True:
            self.gradcam._find_penultimate_output = self.patch


    def __call__(self, image:Image.Image, auto_resize:bool=True, image_size=None):
        #assert check_argument_types()
        gradcam = self.gradcam
        get_loss = self.get_loss
        if auto_resize and image_size is None:
            image_size = self.learner.ds.img_sz_list.get_size()
        if image_size:
            image = image.resize(image_size)

        X = np.asarray(image, np.float32)
        X = self.learner.prewhiten(X)
        X = np.expand_dims(X, 0)

        cam = gradcam(get_loss,
                      X,
                      penultimate_layer=-1, # model.layers number
                      seek_penultimate_conv_layer=True,
                     )
        cam = normalize(cam)
        heatmap = np.uint8(cm.jet(cam[0])[..., :3] * 255)
        plt.imshow(image)
        plt.imshow(heatmap, cmap='jet', alpha=0.5)
        plt.show()


    def patch(self, *args, **kwargs):
        """Path _find_penultimate_output method of tf_keras_vis"""
        if self.learner.include_top:
            return self.learner.model.layers[-1].output
        return self.learner.model.layers[0].get_output_at(-1)


    def model_modifier(self, m):
        """Sets last activation to linear
        """
        m.layers[-1].activation = tf.keras.activations.linear
        return m

    def get_loss(self, preds):
        if self.learner.NUM_CLASSES == 2:
            ret = preds[0]
        else:
            index = tf.argmax(tf.math.softmax(preds), axis=1)[0]
            # print(index, preds.shape)
            ret = preds[0, index]
            print(f'index: {index}')
        return ret


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_datagenerator.ipynb (unless otherwise specified).

__all__ = ['get_filenames', 'get_label', 'ImageSizeList', 'Dataset']

# Cell
import tensorflow as tf
import numpy as np
from glob import glob
import random
import os
import pathlib
from functools import partial
from typing import Union

from .image import read_image, resize_image

# Cell
def get_filenames(root_dir):
        root_dir = pathlib.Path(root_dir)
        return glob(str(root_dir/'*/*'))

def get_label(filename):
    return filename.split('/')[-2]


# Cell
class ImageSizeList():
    def __init__(self, img_sz_list=None):

        if isinstance(img_sz_list, (list, tuple)):
            if len(img_sz_list)!=0 and not isinstance(img_sz_list[0], (list, tuple)):
                img_sz_list = [img_sz_list][:]

        self.start_size = None
        self.last_size = None
        self.curr_size = None
        self.img_sz_list = img_sz_list

        try:
            self.start_size = img_sz_list[0]
            self.last_size = img_sz_list[-1]
            self.curr_size = img_sz_list[0]
        except (IndexError, TypeError) as e:
            print('No item present in the image size list')
            self.curr_size = None # no item present in the list


    def get_size(self):
        img_sz_list = self.img_sz_list
        try:
            self.curr_size = img_sz_list.pop(0)
        except (IndexError, AttributeError) as e:
            print(f'Returning the last set size which is: {self.curr_size}')

        return self.curr_size

# Cell
class Dataset():
    MAPPINGS = {
        'PY_TO_TF': {str:tf.string, int:tf.int32, float:tf.float32},

        }

    def __init__(self, root_dir, image_size=[], transforms=None, default_encode=True, **kwargs):
        self.get_filenames = get_filenames
        self.read_image = read_image
        self.get_label = get_label
        self.transforms = transforms

        self.root_dir = root_dir
        self.default_encode = default_encode
        self.filenames = self.get_filenames(root_dir)
        self.num_files = len(self.filenames)
        self.image_size = image_size
        self.img_sz_list= ImageSizeList(self.image_size[:])

        self.labels = kwargs.get('labels', self.get_labels())


    def __len__(self): return len(self.filenames)


    def _process(self, filename):
        image = self.read_image(filename)
        label = self.get_label(filename)
        return image, label


    def _reload(self):
        self.filenames  = self.get_filenames(self.root_dir)
        self.num_files = len(self.filenames)
        self.img_sz_list = ImageSizeList(None or self.image_size[:])
        self.labels = self.get_labels()

    def _capture_return_types(self):
        return_types = []
        for e in self.generator():
            outputs = e
            break
        if isinstance(outputs, tuple):
            for ret_type in outputs:
                return_types.append(
                    ret_type.dtype if tf.is_tensor(ret_type) else Dataset.MAPPINGS['PY_TO_TF'][type(ret_type)]
                )
        else:
            return_types.append(
                ret_type.dtype if tf.is_tensor(ret_type) else Dataset.MAPPINGS['PY_TO_TF'][type(ret_type)]
            )
        return tuple(return_types)


    def __getitem__(self, idx):
        filename = self.filenames[idx]
        return self._process(filename)

    def update_component(self, component_name, new_component, reload=True):
        setattr(self, component_name, new_component)
        print(f'{component_name} updated with {new_component}')
        self._reload()


    def get_labels(self):
        # get labels should also update self.num_classes
        root_dir = self.root_dir
        labels = set()
        folders = glob(f'{root_dir}/*')
        for folder in folders:
            labels.add(os.path.basename(folder))

        labels = sorted(labels)
        self.NUM_CLASSES = len(labels)
        self.label_to_idx = {label:i for i, label in enumerate(labels)}

        return labels

    def label_encoder(self, label): return self.label_to_idx[label]


    def generator(self, shuffle=False):
        if shuffle: random.shuffle(self.filenames)
        img_sz = self.img_sz_list.get_size()
        n = len(self.filenames)
        for i in range(n):
            image, label = self.__getitem__(i)
            if img_sz: image = resize_image(image, img_sz)
            if self.transforms: image = self.transforms(image)
            if self.default_encode is True:
                label = self.label_encoder(label)
            yield image, label


    def get_tf_dataset(self, output_shape=None, shuffle=True):
        return_types = self._capture_return_types()
        self._reload()
        generator = partial(self.generator, shuffle=shuffle)
        datagen = tf.data.Dataset.from_generator(
            generator,
            return_types,
            output_shape
        )

        return datagen